#include <stdio.h>
#include <dos.h>   /* (Microsoft file) includes for int86() */

/*   VBIOS.C: Various cursor and i/o routine using
 *   the bios interrupts (see below for greater detail):
 *----------------------------------------------------------
 *   Copyright (C) 1987 Allen I. Holub. All rights reserved.
 *----------------------------------------------------------
 * Externally accessable routines:
 *
 * int   vb_getpage ()          Get active video page #
 * void  vb_putchar (c)         write a single character
 * void  vb_getchar (c)         get a key from the bios.
 * void  vb_puts    (s, move)   write a string
 * void  vb_replace (c)         write char w/o moving cursor
 * int   vb_inchar  (attrib)    Get character & attribute
 * 
 * void  vb_setcur  (posn)      Set cur pos as int on cur page
 * int   vb_getcur  ()          Get cur pos as int from cur page
 * void  vb_ctoyx (y,x)         Set cursor position to (y,x)
 * void  vb_getyx (&y, &x)      Get cursor position
 *
 * int   vb_iscolor()           color monitor installed
 * void  vb_cursize (top,bot)   Set cursor size
 * void  vb_blockcur()          make a block cursor
 * void  vb_normalcur()         revert to a normal cursor
 *
 * void  vb_scroll(l,r,t,b,a)   Scroll region
 */

/*--------------------------------------------------------*/

extern int      int86( int, union REGS *, union REGS *);

/*--------------------------------------------------------*/

#define VIDEO_INT     0x10    /* Video interrupt          */
#define KB_INT        0x16    /* Keyboard interrupt       */

#define CUR_SIZE      0x1     /* Set cursor size          */
#define SET_POSN      0x2     /* Modify cursor posn       */
#define READ_POSN     0x3     /* Read current cursor posn */
#define WRITE         0x9     /* Write character          */
#define WRITE_TTY     0xe     /* Write char & move cursor */
#define GET_VMODE     0xf     /* Get video mode & disp pg */

/*--------------------------------------------------------*/

static union REGS       Regs;      /* Used to talk to DOS */
static int              Attribute; /* Current attribute   */

/*--------------------------------------------------------*/

void    vb_scroll( x_left, x_right, y_top, y_bottom, amt )
{
        /* Scroll the indicated region on the screen.
         * If amt is negative, scroll down.
         */

        if( amt < 0 )
        {
            Regs.h.ah = 7    ;
            Regs.h.al = -amt ;
        }
        else
        {
            Regs.h.ah = 6   ;
            Regs.h.al = amt ;
        }

        Regs.h.bh = 0x07    ;
        Regs.h.cl = x_left  ;
        Regs.h.ch = y_top   ;
        Regs.h.dl = x_right ;
        Regs.h.dh = y_bottom ;
        int86(0x10, &Regs, &Regs);
}

/*--------------------------------------------------------*/

int     vb_inchar( attrib )
int     *attrib;
{
        /* Return the character at the current cursor 
         * position and, if attrib is non-NULL, put the
         * attribute there. Note that vb_getpage() will mess
         * up the fields in the Regs structure so it must
         * be called first.
         */

        Regs.h.bh = vb_getpage() ;
        Regs.h.ah = 8 ;

        int86( VIDEO_INT, &Regs, &Regs );

        if( attrib )
                *attrib = Regs.h.ah & 0xff ;

        return( Regs.h.al & 0xff );
}

/*--------------------------------------------------------*/

int     vb_getpage()
{
        /* Returns the currently active display page number
         */

        Regs.h.ah = GET_VMODE;
        int86( VIDEO_INT, &Regs, &Regs );

        return (int) Regs.h.bh ;        
}

/*--------------------------------------------------------*/

void    vb_cursize( top_line, bot_line )
{
        /* Scan lines are numberd 0 at the top and 7 at the
         * bottom on the color card. On the monochrome card
         * they're 0-12. If top & bot are reversed you'll
         * get a 2 part cursor. Top_line determines the
         * position of the top scan line of the cursor,
         * bot_line is the bottom. A normal cursor can be
         * created with vb_cursize(6,7). Cursize(0,7) will
         * fill the entire area occupied by a character.
         * Cursize(0,1) will put a line over the character
         * rather than under it.
         */

        Regs.h.ch = top_line ;
        Regs.h.cl = bot_line ;
        Regs.h.ah = CUR_SIZE ;
        int86( VIDEO_INT, &Regs, &Regs );
}

/*--------------------------------------------------------*/

int vb_iscolor()   /* Returns true if a color card is active */
{ 
        Regs.h.ah = GET_VMODE ;
        int86( VIDEO_INT, &Regs, &Regs );
        return( Regs.h.al != 7 );
}

void vb_blockcur()     /* Make the cursor a block curser */
{
        vb_cursize( 0, vb_iscolor() ? 7 : 12 );
}

void vb_normalcur()      /* Make it an underline cursor */
{
        if( vb_iscolor() )
                vb_cursize(  6,  7 );
        else
                vb_cursize( 11, 12 );
}

/*--------------------------------------------------------*/

void    vb_setcur( posn )
int     posn;
{
        /* Modify current cursor position. The top byte of
         * "posn" value holds the row (y), the bottom byte,
         * the column (x). The top-left corner of the screen
         * is (0,0). Pagenum is the video page number. Note
         * that vb_getpage() will mess up the fields in the
         * Regs structure so it must be called first.
         */

        Regs.h.bh = vb_getpage() ;
        Regs.x.dx = posn         ;
        Regs.h.ah = SET_POSN     ;
        int86( VIDEO_INT, &Regs, &Regs );
}

int     vb_getcur()
{
        /* Get current cursor position. The top byte of the
         * return value holds the row, the bottom by the
         * column. Pagenum is the video page number. Note
         * that vb_getpage() will mess up the fields in the
         * Regs structure so it must be called first.
         */

        Regs.h.bh = vb_getpage() ;
        Regs.h.ah = READ_POSN    ;
        int86( VIDEO_INT, &Regs, &Regs );
        return( Regs.x.dx );
}

/*----------------------------------------------------------
 * vb_cotyx() and vb_getyx also get the cursor position.
 * They use x and y values, however.
 */

void    vb_ctoyx ( y, x )
{
        vb_setcur( (y << 8) | (x & 0xff) );
}

void    vb_getyx( yp, xp )
int     *yp, *xp;
{
        register int    posn;

        posn = vb_getcur();
        *xp  = posn        & 0xff ;
        *yp  = (posn >> 8) & 0xff ;
}

/*--------------------------------------------------------*/

vb_replace(c)
{
        /* Overwrite the character at the current cursor
         * position without moving the cursor.
         */

        Regs.h.ah = 10 ;
        Regs.h.al = c;          /* write c              */
        Regs.h.bl = 0x07;       /* Normal characters    */
        Regs.h.bh = 0;          /* Display page 0       */
        Regs.x.cx = 1;          /* # of times to write  */

        int86( VIDEO_INT, &Regs, &Regs );
}

/*--------------------------------------------------------*/

vb_putchar( c )
{
        /* Write a character to the screen in TTY mode.
         * Only normal printing characters, BS, BEL, CR and
         * LF are recognized. The cursor is automatically
         * advanced and lines will wrap.
         */

        Regs.h.bl = 0x07;
        Regs.h.al = c;
        Regs.h.ah = WRITE_TTY ;
        int86( VIDEO_INT, &Regs, &Regs );
}

/*--------------------------------------------------------*/

vb_puts( str, move_cur )
register char   *str;
{
        /* Write a string to the screen in TTY mode. If
         * move_cur is true the cursor is left at the end
         * of string. If not the cursor will be restored to
         * its original position (before the write).
         */

        register int    posn;

        if( !move_cur )
                posn = vb_getcur();

        while( *str )
                vb_putchar( *str++ );

        if( !move_cur )
                vb_setcur( posn );
}

/*--------------------------------------------------------*/

int     vb_getchar()
{
        /* Get a character with a direct video bios call.
         * This routine can be used to complement stderr as
         * it can be used to get characters from the keyboard,
         * even when input redirected. The typed character
         * is returned in the low byte of the returned
         * integer, the high byte holds the auxillary byte
         * used to mark ALT keys and such. See the Technical
         * Reference for more info.
         */

        Regs.h.ah = 0 ;
        int86( KB_INT, &Regs, &Regs );
        return( (int)Regs.x.ax );
}

/*--------------------------------------------------------*/
#ifdef MAIN

main()
{
        vb_replace( 'X' );
        vb_putchar('\n');
        vb_putchar('\r');
}

#endif
